# p5.js Friendly Error System (FES)

## Overview

The Friendly Error System (FES) is a system designed to help new programmers with common user errors as they get started learning. It catches common beginning errors and provides clear language and links to help a user resolve the error. FES is only applied to functions that are ones a user might encounter when they are just starting. An exception is made for particular common gotchas such as trying to load files without running a server, or calling loadImage() with a URL, etc.

The goal is to create accessible error messages to supplement the often cryptic console errors. For example, Javascript has no support for type checking as default making errors in parameter entry was harder to detect for new Javascript developers.

Messages generated by FES are written in natural language, linked to documentation, and assume a beginner level. The errors are triggered in multiple files through out p5.js, but most of the work and error writing happens in `src/core/error_helpers.js`.  

By default, FES is enabled for `p5.js`, whereas completely disabled in `p5.min.js`. It is possible to disable FES by setting `p5.disableFriendlyErrors = true;`.

So far FES is able to detect and print messages for four kinds of errors: 

1. `validateParameters()` checks a function’s input parameters based on inline documentations

2. `friendlyFileLoadError()` catches file loading errors. These two kinds of error checking have been integrated into an existing (selected set of) p5 functions, but developers can add them to more p5 functions, or their own libraries, by calling `p5._validateParameters()`

3. `friendlyError()` can be called by any function to offer a helpful error message

4. `helpForMisusedAtTopLevelCode()` is called on window load to check for use of p5.js functions outside of setup() or draw()

Please also see inline notes in [src/core/error_helper.js](https://github.com/processing/p5.js/blob/master/src/core/error_helpers.js) for more technical information.

### `core/error_helpers/friendlyFileLoadError()`: 
* This function generates and displays friendly error messages if a file fails to load correctly. It also checks if the size of a file might be too large to load and produces a warning. 
* This can be called through : `p5._friendlyFileLoadError(ERROR_TYPE, FILE_PATH)`.
* file loading error example:
````javascript
/// missing font file
let myFont;
function preload() {
  myFont = loadFont('assets/OpenSans-Regular.ttf');
};
function setup() {
  fill('#ED225D');
  textFont(myFont);
  textSize(36);
  text('p5*js', 10, 50);
};
function draw() {};
/// FES will generate the following message in the console:
/// > p5.js says: It looks like there was a problem loading your font. Try checking if the file path [assets/OpenSans-Regular.ttf] is correct, hosting the font online, or running a local server.[https://github.com/processing/p5.js/wiki/Local-server]
````
* Currently version contains templates for generating error messages for `image`, `XML`, `table`, `text`, `json` and `font` files.
* Implemented to `image/loading_displaying/loadImage()`, `io/files/loadFont()`, `io/files/loadTable()`, `io/files/loadJSON()`, `io/files/loadStrings()`, `io/files/loadXML()`, `io/files/loadBytes()`.
* Error while loading a file due to its large size is implemented to all loadX methods.

### `core/error_helpers/validateParameters()`:
* This function runs parameter validation by matching the input parameters with information from `docs/reference/data.json`, which is created from the function's inline documentation. It checks that a function call contains the correct number and the correct type of parameters. 
* missing parameter example:
````javascript
arc(1, 1, 10.5, 10);
/// FES will generate the following message in the console:
/// > p5.js says: It looks like arc() received an empty variable in spot #4 (zero-based index). If not intentional, this is often a problem with scope: [https://p5js.org/examples/data-variable-scope.html]. [http://p5js.org/reference/#p5/arc]
/// > p5.js says: It looks like arc() received an empty variable in spot #5 (zero-based index). If not intentional, this is often a problem with scope: [https://p5js.org/examples/data-variable-scope.html]. [http://p5js.org/reference/#p5/arc]

````
* wrong type example:
````javascript
arc('1', 1, 10.5, 10, 0, Math.PI, 'pie');
/// FES will generate the following message in the console:
/// > p5.js says: arc() was expecting Number for parameter #0 (zero-based index), received string instead. [http://p5js.org/reference/#p5/arc]
````
* This can be called through: `p5._validateParameters(FUNCT_NAME, ARGUMENTS)` 
or, `p5.prototype._validateParameters(FUNCT_NAME, ARGUMENTS)` inside the function that requires parameter validation. It is recommended to use static version, `p5._validateParameters` for general purposes. `p5.prototype._validateParameters(FUNCT_NAME, ARGUMENTS)` mainly remained for debugging and unit testing purposes.
* Implemented to functions in `color/creating_reading`, `core/2d_primitives`, `core/curves`, and `utilities/string_functions`. 

## Additional Features
* The FES welcomes the developer to p5 and the friendly debugger. 
* The FES works in the IDE and the web editor. 

## Notes for Developers
* When creating p5.js Objects: any p5.js objects that will be used for parameters will need to assign value for `name` parameter (name of the object) within the class declaration. e.g.: 
````javascript 
p5.newObject = function(parameter) {
   this.parameter = 'some parameter';
   this.name = 'p5.newObject';
};
````
* Inline documentation: allowed parameter types are `Boolean`, `Number`, `String`, and name of the object (see the above bullet point). Use `Array` for any types of Array parameters. If needed, explain what kind of the specific types of array parameter are allowed (e.g. `Number[]`, `String[]`) in the description section.
* Currently supported class types (have their `name` parameter): `p5.Color`, `p5.Element`, `p5.Graphics`, `p5.Renderer`, `p5.Renderer2D`, `p5.Image`, `p5.Table`, `p5.TableRow`, `p5.XML`, `p5.Vector`, `p5.Font`, `p5.Geometry`, `p5.Matrix`, `p5.RendererGL`.

## Disable the FES

By default, FES is enabled for p5.js, and disabled in p5.min.js to prevent FES functions slowing down the process. The error checking system can significantly slow down your code (up to ~10x in some cases). See the [friendly error performance test](https://github.com/processing/p5.js-website/blob/master/dist/assets/learn/performance/code/friendly-error-system/).

You can disable this with one line of code at the top of your sketch:

```javascript
p5.disableFriendlyErrors = true; // disables FES

function setup() {
  // Do setup stuff
}

function draw() {
  // Do drawing stuff
}
```

Note that this will disable the parts of the FES that cause performance slowdown (like argument checking). Friendly errors that have no performance cost (like giving an descriptive error if a file load fails, or warning you if you try to override p5.js functions in the global space), will remain in place. 

## Known Limitations
* The friendly error system slows the program down, so there is an option to turn it off via setting `p5.disableFriendlyErrors = true;`. In addition, the friendly error system is omitted by default in the minified (`p5.min.js`) version.
* FES may still result in false negative cases. These are usually caused by the mismatch between designs of the  functions (e.g. drawing functions those are originally designed to be used interchangeably in both 2D and 3D settings) with actual usage cases. For example, drawing a 3D line with 
```javascript 
const x3; // undefined
line(0, 0, 100, 100, x3, Math.PI);
```
 will escape FES, because there is an acceptable parameter pattern (`Number`, `Number`, `Number`, `Number`) in `line()`'s inline documentation for drawing in 2D setting. This also means the current version of FES doesn't check for the environmental variables such as `_renderer.isP3D`.
 * FES is only able to detect global variables overwritten when declared using `const` or `var`. If `let` is used, they go undetected. This is not currently solvable due to the way `let` instantiates variables.

## In The Works
* Identify more common error types and generalize with FES (e.g. `bezierVertex()`, `quadraticVertex()` - required object not initiated; checking Number parameters positive for `nf()` `nfc()` `nfp()` `nfs()`)

## Thoughts for the Future
* re-introduce color coding for the Web Editor.
* More unit testings.
* More intuitive and narrowed down output messages.
* Having a Spanish translation available.
* All the colors are checked for being color blind friendly.
* More elaborate ascii is always welcome! 
* Global Error catching. It would be very helpful to catch the errors the browser is throwing to the console, so we can match them up with friendly comments. 

```javascript
// this snippet wraps window.console methods with a new function to modify their functionality
// it is not currently implemented, but could be to give nicer formatted error messages
const original = window.console;
const original_functions  = {
  log: original.log,
  warn:  original.warn,
  error: original.error
}

["log", "warn", "error"].forEach(function(func){
window.console[func] = function(msg) {
//do something with the msg caught by the wrapper function, then call the original function
original_functions[func].apply(original, arguments)
};
});
```
